#include "i2c.hpp"

using namespace stm32f072xb;

I2c::I2c(I2C_TypeDef* type) : _type(type) {
	setupI2c();
	setupGpio();
	softwareReset();
}

void I2c::configureI2cClock() {
	// Clear TIMINGR
	_type->TIMINGR = 0;

	// I2CCLK is at 8MHz
	SET_BIT(_type->TIMINGR, (0x1 << I2C_TIMINGR_PRESC_Pos));
	SET_BIT(_type->TIMINGR, (0x13 << I2C_TIMINGR_SCLL_Pos));
	SET_BIT(_type->TIMINGR, (0xF << I2C_TIMINGR_SCLH_Pos));
	SET_BIT(_type->TIMINGR, (0x4 << I2C_TIMINGR_SCLDEL_Pos));
	SET_BIT(_type->TIMINGR, (0x2 << I2C_TIMINGR_SDADEL_Pos));
}

void I2c::setupGpio() {
	if (_type == I2C1) {
		// Setup GPIO - Open Drain
		I2cGpioInit sdaInit = {GpioPin::P9, Pupdr::PULL_UP, AlternateFunction::AF1};
		I2cGpio<Port::B> sda(sdaInit);
		I2cGpioInit sclInit = {GpioPin::P8, Pupdr::PULL_UP, AlternateFunction::AF1};
		I2cGpio<Port::B> scl(sclInit);
	} else if (_type == I2C2) {
		// Setup GPIO - Open Drain
		I2cGpioInit sdaInit = {GpioPin::P11, Pupdr::PULL_UP, AlternateFunction::AF1};
		I2cGpio<Port::B> sda(sdaInit);
		I2cGpioInit sclInit = {GpioPin::P10, Pupdr::PULL_UP, AlternateFunction::AF1};
		I2cGpio<Port::B> scl(sclInit);
	}
}

void I2c::enableI2cClock() {
	if (_type == I2C1) {
		// Set HSI (8MHz) as I2C1 clock (Only available for I2C1)
		CLEAR_BIT(RCC->CFGR3, RCC_CFGR3_I2C1SW);

		// Enable I2C peripheral clock
		SET_BIT(RCC->APB1ENR, RCC_APB1ENR_I2C1EN);
	} else if (_type == I2C2) {
		// Enable I2C peripheral clock
		SET_BIT(RCC->APB1ENR, RCC_APB1ENR_I2C2EN);
	}
}

void I2c::setupI2c() {
	// Chose clock source and enable clock
	enableI2cClock();

	// Make sure I2C is disabled
	CLEAR_BIT(_type->CR1, I2C_CR1_PE);

	// Setup timing 8MHz clock
	configureI2cClock();

	// Setup 7 bit addressing
	CLEAR_BIT(_type->CR2, I2C_CR2_ADD10);

	// I2C automatically sends stop bit after nBytes have been transmitted
	// I2C automatically sends NACKF before stop bit after nBytes have been received in master receive mode
	SET_BIT(_type->CR2, I2C_CR2_AUTOEND);

	// Enable clock stretching (This needs to be enabled as we are in master mode)
	CLEAR_BIT(_type->CR1, I2C_CR1_NOSTRETCH);

	// Enable analog noise filter
	CLEAR_BIT(_type->CR1, I2C_CR1_ANFOFF);

	// Enable I2C peripheral
	SET_BIT(_type->CR1, I2C_CR1_PE);
}

void I2c::setupTransmission(uint8_t slaveAddress, uint8_t bufferSize) {
	// Clear Start bit to allow for changes
	CLEAR_BIT(_type->CR2, I2C_CR2_START);

	// Setup SADD and NBYTES
	CLEAR_BIT(_type->CR2, I2C_CR2_NBYTES);
	SET_BIT(_type->CR2, bufferSize << I2C_CR2_NBYTES_Pos);
	CLEAR_BIT(_type->CR2, I2C_CR2_SADD);
	SET_BIT(_type->CR2, slaveAddress << I2C_CR2_SADD_Pos);
}

void I2c::blockingWrite(uint8_t slaveAddress, uint8_t* bytes, uint8_t bufferSize) {
	// Send message until it is accepted by the slave
	do {
		// Clear NACK ISR bit
		SET_BIT(_type->ICR, I2C_ICR_NACKCF);

		// Set TXE bit to flush previous content of TXDR
		SET_BIT(_type->ISR, I2C_ISR_TXE);

		// Setup slave address and nBytes to be sent
		setupTransmission(slaveAddress, bufferSize);

		// Clear RD_WRN bit to indicate a write
		CLEAR_BIT(_type->CR2, I2C_CR2_RD_WRN);

		for (int i = 0; i < bufferSize; i++) {
			_type->TXDR = bytes[i];

			if (i == 0)
				SET_BIT(_type->CR2, I2C_CR2_START);

			// TODO - Replace polling with DMA
			while ((_type->ISR & I2C_ISR_TXE) != I2C_ISR_TXE) {
				// If we receive a NACK flush TXDR and stop sending bytes
				if ((_type->ISR & I2C_ISR_NACKF) == I2C_ISR_NACKF) {
					i = bufferSize;
					SET_BIT(_type->ISR, I2C_ISR_TXE);
				}
			}
		}
	} while ((_type->ISR & I2C_ISR_NACKF) == I2C_ISR_NACKF);
}

void I2c::blockingRead(uint8_t slaveAddress, uint8_t* bytes, uint8_t bufferSize) {
	// TODO - make sure All bytes are received

	// Clear NACK ISR bit
	SET_BIT(_type->ICR, I2C_ICR_NACKCF);

	setupTransmission(slaveAddress, bufferSize);

	// Clear RD_WRN bit to indicate a write
	SET_BIT(_type->CR2, I2C_CR2_RD_WRN);

	// Send Start condition to begin receiving data
	SET_BIT(_type->CR2, I2C_CR2_START);

	for (int i = 0; i < bufferSize; i++) {
		while ((_type->ISR & I2C_ISR_RXNE) != I2C_ISR_RXNE)
			;
		bytes[i] = _type->RXDR;
	}
}

void I2c::write(uint8_t slaveAddress, uint8_t bufferSize) {
	setupTransmission(slaveAddress, bufferSize);
	// Clear RD_WRN bit to indicate a write
	CLEAR_BIT(_type->CR2, I2C_CR2_RD_WRN);
	// Send Start condition to begin receiving data
	SET_BIT(_type->CR2, I2C_CR2_START);
}

void I2c::read(uint8_t slaveAddress, uint8_t bufferSize) {
	setupTransmission(slaveAddress, bufferSize);
	// Clear RD_WRN bit to indicate a write
	SET_BIT(_type->CR2, I2C_CR2_RD_WRN);
	// Send Start condition to begin receiving data
	SET_BIT(_type->CR2, I2C_CR2_START);
}

void I2c::softwareReset() {
	CLEAR_BIT(I2C1->CR1, I2C_CR1_PE);
	READ_BIT(I2C1->CR1, I2C_CR1_PE);
	SET_BIT(I2C1->CR1, I2C_CR1_PE);
}

void I2c::enableDmaReception() {
	SET_BIT(_type->CR1, I2C_CR1_RXDMAEN);
}

void I2c::enableDmaTransmission() {
	SET_BIT(_type->CR1, I2C_CR1_TXDMAEN);
}
